//========================================================================================
// main.java
// ---------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
// $Id: //CBT-depot/CompBioToolsApps/TemplateViewer/NativeLoader.java#1 $
//----------------------------------------------------------------------------------------
//
//========================================================================================
//.......1.........2.........3.........4.........5.........6.........7.........8.........9
//
//
//package edu.bcm.atlas.MultiPlatformSupport;

import javax.swing.*;
import javax.swing.event.MouseInputAdapter;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import java.net.*;

//---------------------------------------------------------------------------------------
// NativeLoader 
//---------------------------------------------------------------------------------------
/** 
    Support for loading platform-specific native libraries. <br><br>
    
    <b><i>IMPORTANT NOTE:</i><b> Native loading must be done before any attempt to use 
    members from the native library.  The best way to ensure that is to load the library 
    from a static initializer block. 
*/
public class NativeLoader{
	

		//------------------------------------
		// GetOS 
		//------------------------------------
		/**
			 Gets the OS string from system properties. 

			 The major strings of interest: <br> <br> 
			 Mac OS X, Linux, SunOS  <br> <br>
			 Note that SunOS is what Solaris returns.  
     
			 @param message a <code>String</code> value
		*/
		public String GetOS(){
				// Get all system properties
				Properties props = System.getProperties();
				String propValue = (String)props.get("os.name");
				return(propValue);
		}

		// Eventually I'll have to pay attention to os.arch as well.  
		// Currently os.arch values of interest to me include:
		//
		// sparc
		// i386
		// ppc 
		// 
  
		//------------------------------------
		// * LoadLibrary
		//------------------------------------
		/** 
				Loads a native library from a .jar file. 

				Determines the OS of the system, extracts the appropriate native library 
				from the .jar into a temporary file, then loads that library.  <br><br>

				<b>Note:</b> Assumes libraries are in the .jar file with the following names:<br><br>
				<table border=1> 
				<TR><TH>OSX<TD>libRootName.jnilib
				<TR><TH>Solaris<TD>libRootName.solaris.so 
				<TR><TH> Linux<TD>libRootName.linux.so 
				</table> 
    
				@param RootName the name of the native library without system specific prefix or suffix. 
		*/
		public void LoadLibrary(String RootName){
				String os = GetOS();
				String ResourceName="",TempResourceRoot="",TempResourceSuffix="";
				if (os.equals("Mac OS X")){
						ResourceName = "/lib"+RootName+".jnilib";
						TempResourceRoot="temp"+RootName;
						TempResourceSuffix = ".jnilib";
				}else if (os.equals("Linux")){
						// for now, assume Linux means linux i386... 
						ResourceName = "/lib"+RootName+".linux.so";
						TempResourceRoot="temp"+RootName;
						TempResourceSuffix = ".linux.so";
				}else if (os.equals("SunOS")){
						ResourceName = "/lib"+RootName+".solaris.so";
						TempResourceRoot="temp"+RootName;
						TempResourceSuffix = ".solaris.so";
				}else{
						System.out.println("Unrecognized OS: "+os);
						System.out.println("This software uses native libraries.  Libraries for "+os+" not available.");
						System.exit(1);
				}
				// ... and so on for other OSs...


				//System.out.println("NativeLoader RootName="+RootName+" os="+os+" ResourceName="+ResourceName);


				try{
						URL ResourceURL = NativeLoader.class.getResource(ResourceName);
						//	System.out.println("URL: "+ResourceURL.toString());
						InputStream inputStream = ResourceURL.openStream();
						File temporaryLibrary = File.createTempFile(TempResourceRoot,TempResourceSuffix);
						System.out.println("A temporary file has been created for the native library at:");
						System.out.println("\t"+temporaryLibrary.toString());
						System.out.println("This file will be deleted automatically on normal exit. ");
						System.out.println("If this program exits abnormally, you may want to delete this file manually.");
						System.out.println("\n");
						
						FileOutputStream  outStream = new FileOutputStream(temporaryLibrary);
						byte[] array = new byte[8192];
						for(int i = inputStream.read(array); i != -1; i=inputStream.read(array)){
								outStream.write(array,0,i);
						}
						outStream.close();
						// Make sure to clean it up when we are done. 
						temporaryLibrary.deleteOnExit();      
						try {
								System.load(temporaryLibrary.getPath());
						} catch (UnsatisfiedLinkError e) {
								System.err.println("Native code library failed to load from jar file.");
								System.err.println("Message: "+e.getMessage());
						}
				}catch (IOException ie){
						System.out.println("Unable to get Resource.");
				}    
		}		
}

















