#include "ReadPainter.hh"

bool cmpByPadLoc(padInfo a,padInfo b)
{
  return(a.mPadLoc<b.mPadLoc);
}

bool cmpByReadIdx(padInfo a,padInfo b)
{
  return(a.mReadIdx<b.mReadIdx);
}

//----------------------------------------
// * GetNumReadsInWindow
//----------------------------------------
// 
int 
ReadPainter::GetNumReadsInWindow(int winStart,int winEnd)
{
  string outString;
  int rSize = mMergedReads.size();

  // Examine each read to see if it is in our window and count it.  
  int numReadsInWin = 0;
  for(int r = 0;r< rSize;r++){
    for(int c = winStart;c<winEnd;c++){				
      int rIdx = SIdx2RIdx(c,r);
      
      if ((rIdx > 0) && (rIdx < mMergedReads[r].size())){
	numReadsInWin++;
	break;
      }
    }
  }
  return(numReadsInWin);
}

//----------------------------------------
// * GetWindowReadName
//----------------------------------------
// Returns the rIdx'th read in the defined window.  
const char* 
ReadPainter::GetWindowReadName(int winStart,int winEnd,int targetIdx)
{
  string outString;
  int rSize = mMergedReads.size();
  
  IntSet WinReads;
  
  // Examine each read to see if it is in our window and if so 
  // if it's the rth such read.  
  int readCount = 0;  
  for(int r = 0;r< rSize;r++){
    for(int c = winStart;c<winEnd;c++){				
      int rIdx = SIdx2RIdx(c,r);
      if ((rIdx > 0) && (rIdx < mMergedReads[r].size())){
	if (readCount == targetIdx) return(mMergedReads[r].mName.c_str());
	readCount++; // Found another read in our window...
	break;
      }
    }
  }
  return("NOT FOUND");
}

char 
ReadPainter::GetSourceBase(int cIdx){
  return(mUnionSource[cIdx]);
};



//----------------------------------------
// * Printalignment
//----------------------------------------
//
void
ReadPainter::PrintAlignment(ostream &fout,int start,int end,int lineSize)
{
  if (end < 0) end = mUnionSource.size() -1;

  int c = 0;
  int lineStart = start;
  do{
    //Print consensus
    PrintSpacePaddedString(fout,20,mUnionSource.mName);
    PrintSpacePaddedNum(fout,10,lineStart);
    fout<<"  ";
    //    fout.width(7);
    //fout<<lineStart<<" ";
    for(int idx = lineStart;idx< lineStart+lineSize;idx++){
      if ((idx >= 0) && (idx < mUnionSource.size())){
	fout<<mUnionSource[idx];
      }else if (idx > mUnionSource.size()){
	fout<<".";
      }else{
	fout<<"%";  // How did we get here??
      }
    }
    fout<<endl;

    // Print each of the reads. 
    for(int r = 0;r< mMergedReads.size();r++){
      int readStart = SIdx2RIdx(0,r);
      int readLineStart = readStart + lineStart;

      // Skip reads which are not in this region. 
      if (readLineStart < 0) continue;

      PrintSpacePaddedString(fout,20,mMergedReads[r].mName);
      PrintSpacePaddedNum(fout,10,readLineStart);
      fout<<"  ";
      //      fout.width(7);
      //fout<<readLineStart<<" ";
      for(int idx = lineStart;idx< lineStart+lineSize;idx++){
	int readIdx = readStart + idx;
	if ((readIdx >= 0) && (readIdx < mMergedReads[r].size())){
	  fout<<mMergedReads[r][readIdx];
	}else if (readIdx >= mMergedReads[r].size()){
	  fout<<".";
	}else{
	  fout<<"%"; // How did we get here??
	}
      }
      fout<<endl;
    }  	 
    fout<<endl;
    lineStart+=lineSize;
  }while((lineStart < mUnionSource.size()) && (lineStart <=end));
}


//----------------------------------------
// * GetAlignedCharacters
//----------------------------------------
// Returns a vertical stripe of read bases at contig position cIdx
// winStart and winEnd are used to determine how many reads to pad 
// with spaces.  That is, there may be reads in the window that do 
// not appear at cIdx, but vertical padding must reflect the whole 
// window's worth of characters.  
// 
// ........ACTGAGATTACTAGGATAC..................
// ..ACATTAACTGAGATTACTGGGATAGGGCAG.............
// ................................AAGGTACTGGGGG
const char*
ReadPainter::GetAlignedCharacters(int winStart,int winEnd,int cIdx)
{
  string outString;
  int rSize = mMergedReads.size();
  
  // Examine each read to see if it is in our window and if so 
  // if it is also in our column.  If in our window we add a pad
  // character to our output.  If in column we add the read character
  // to out output. 
  int numReadsInWin = 0;
  for(int r = 0;r< rSize;r++){
    for(int c = winStart;c<winEnd;c++){				
      int rIdx = SIdx2RIdx(c,r);
      if ((rIdx >= 0) && (rIdx < mMergedReads[r].size())){
	
	// Read is in the window, see if it is also in the column
	int columnIdx = SIdx2RIdx(cIdx,r);
	if ((columnIdx >= 0) && (columnIdx < mMergedReads[r].size())){
	  outString+=mMergedReads[r][columnIdx];
	}else{
	  outString+=" ";
	}
	// Either way, we're done with this read. 
	break;
      }
    }
  }
  return(outString.c_str());
}

int 
ReadPainter::GetNumReads()
{
  return(mMergedReads.size());
}

const char*
ReadPainter::GetReadName(int rIdx)
{
  return(mMergedReads[rIdx].mName.c_str());
}

int 
ReadPainter::GetSourceSize()
{
  // Initially, there may not be a Union Source. 
  if (mUnionSource.size() == 0){
    return(mSource.size());
  }else{
    return(mUnionSource.size());
  }
}


void 
ReadPainter::ReadSource(char* sName)
{
  string sNameStr = sName;
  ReadSource(sNameStr);
}

void 
ReadPainter::ReadCM(char* CMName){
  string CMNameStr = CMName;
  ReadCM(CMNameStr);	
}

void 
ReadPainter::MergeAlignments()
{
  cout<<"ConstructSource..."<<flush;
  ConstructSourceWithUnionOfReadPads();
	
  FAFile FA;
  FA.Write("usrctest.fa", mUnionSource);

  cout<<"Init With Queries ..."<<flush;
  InitWithQueries();
  cout<<"MergePads..."<<flush;
  MergeReadPads(); 
}


// This must be failing???  

int 
ReadPainter::GetMaxPadIdx(int & padIdx,vector<padInfo> &padLocs){
  
  int targetLoc = padLocs[padIdx].mPadLoc;
  int currentRead = padLocs[padIdx].mReadIdx;

  int maxPadCount = 0;
  do{
    int padCount = 0;

    // Quit if we've reached another pad or run out of reads. 
    if (padIdx >= padLocs.size()) break;
    if (padLocs[padIdx].mPadLoc != targetLoc) break;

    do{
      // Quit if we've reached another pad or run out of reads. 
      if (padIdx >= padLocs.size()) break;
      if (padLocs[padIdx].mPadLoc != targetLoc) break;

      padCount++;
      padIdx++;

    }while(padLocs[padIdx].mReadIdx == currentRead);

    currentRead = padLocs[padIdx].mReadIdx;
    if (padCount > maxPadCount) maxPadCount = padCount;	 

  }while(padLocs[padIdx].mPadLoc == targetLoc);
	

  return(maxPadCount);
}

void 
ReadPainter::GetSortedPadLocs(CMParse &CM,vector<padInfo> &padLocs){

  int padCount = 0;
  //vector<int> padLocs;

  for(int i = 0;i < CM.size();i++){

    string bugQName = CM[i].mQuerySeq.mName;

    int padNum = 0;
    for(int j = 0;j< CM[i].mSourceSeq.size();j++){
      if (CM[i].mSourceSeq[j] == '-'){
	int sourceLoc = CM[i].mSourceStart+j;
	// Want to point to the base before the pad, adjusted 
	// back to the original source coordinates.  
	int adjustedLoc = sourceLoc - padNum-1;
			
	padInfo temp;
	temp.mPadLoc = adjustedLoc;
	temp.mReadIdx = i;

	padLocs.push_back(temp);
	padNum++;
	padCount++;
      }
    }
  }

  stable_sort(padLocs.begin(),padLocs.end(),cmpByReadIdx);	
  stable_sort(padLocs.begin(),padLocs.end(),cmpByPadLoc);

  // KJD DEBUG
  /*
    for(int i = 0;i < padLocs.size();i++){
    if ((padLocs[i].mPadLoc > 2000) && (padLocs[i].mPadLoc < 2600)){
    cout<<"padLoc: "<<padLocs[i].mPadLoc<<"\t"<<padLocs[i].mReadIdx<<endl;
    }
    }
  */
}

//void ConstructSourceWithUnionOfReadPads2(CMParse &CM,
//CMParse &CM,
//Sequence &Source,
//Sequence &NewSource)
void 
ReadPainter::ConstructSourceWithUnionOfReadPads()
{ 
  vector<padInfo> padLocs;
  GetSortedPadLocs(mCM,padLocs);

  mUnionSource.mName = "UnionSource";

	
  int padIdx = 0;
  for(int i = 0;i < mSource.size();i++){
    if (padIdx < padLocs.size()){
      if (i < padLocs[padIdx].mPadLoc){
	mUnionSource.push_back(mSource[i]);
      }else{
	if (i == padLocs[padIdx].mPadLoc){
	  // Found a pad.   Add the source base before the pad. 
	  mUnionSource.push_back(mSource[i]);
			 
	  // Now this pad  may be part of a run of pads, 
	  // so we need to find, for all reads with pads at this location, 
	  // the read with the maximum number of pads.   Then we need to 
	  // add that many pads to this location, and skip over all the 
	  // other pads here.  This function will automatically update 
	  // padIdx to point to the next pad after this block.			 
	  int numPadsAtLoc = GetMaxPadIdx(padIdx, padLocs);		
	 
	  for(int p = 0; p < numPadsAtLoc;p++){				
	    mUnionSource.push_back('-');
	  }				
		  
	  // I want to keep mSourceStart relative to the -padded- sequence, so 
	  // every time I add a pad, I will need to increment the start of all reads 
	  // with start coordinates that occur after this pad.   This is not especially 
	  //efficient way to do this, but hey, there aren't that many pads...  
	  // Since I may have added a block of pads, need to increment source start by 
	  // that number of pads.  
	  //			 for(int r = 0;r< CM.size();r++){
	  //		if (CM[r].mSourceStart > i) CM[r].mSourceStart+=numPadsAtLoc;
	  // }
	}else{
	  cerr<<"??? i = "<<i<<" padIdx="<<padIdx<<" padLocs[padIdx]=";
	  cerr<<padLocs[padIdx].mPadLoc<<endl;			
	}
      }
    }else{
      // Seen all the pads, now just add in sequence...
      mUnionSource.push_back(mSource[i]);
    }
  }
}




void 
ReadPainter::InitWithQueries()
{
  mBaseLocs.resize(mCM.size());

  for(int i = 0;i < mCM.size();i++){
    mAlignedReads.push_back(mCM[i].mQuerySeq);

    // Since CMParse converts this to zero base, we need to convert it back 
    // to one based.  
    // KJD Once CMParse is fixed to preserve CM coordinates, this will need
    // to be changed back. 
    int startIdx = mCM[i].mSourceStart+1;
    mOffsets.push_back(startIdx);

    int idx = startIdx -1;
    for(int j = 0;j < mCM[i].mQuerySeq.size();j++){		    
      // For non-pads, put idx into BaseLocations and increment idx
      if ((mCM[i].mSourceSeq[j] != '*') && 
	  (mCM[i].mSourceSeq[j] != '-')){
	idx++;
	mBaseLocs[i].push_back(idx);
      }else{
	// When there is a pad, just put idx into BaseLocations, but don't increment. 
	mBaseLocs[i].push_back(idx);
      }
    }
  }
}


// The original match to the SyntheticConsensus will look like:
// 
// 	     12234567890123
// Con   AC*CGTTCATCATG
// R1    ACTC*TTCATCAT
//
// Each base in R1 has a Depadded coordinate relative to Con. 
//
// Now, once we've created a Union consensus with all of the 
// read pads, we have something like:
// 
//      12233345678990123
// UCon AC*C**GTTCAT*CATG
//
// So, the next step is to match each read base to the consensus
// position number it had originally above. We match depadded coordinate 
// to depadded coordinate.  That is, the depadded coordinate for each base 
// should be the same after expanding some of the pads runs in UnionCon. 
// For example:
//
//      12233345678990123   Depadded SU Coordinates 
//      01234567890123456   Padded SU Coordinates 
// UCon AC*C**GTTCAT*CATG
// R1   ACTC  *TTCAT CAT
// 
//      1223  456789 012    BaseLocs
//
//      0123456789012345    readIdx
//
// 
// !! Sequences which begin or end in pads are verboten!!
// 

void 
ReadPainter::MergeReadPads()
{

  for(int i = 0;i < mAlignedReads.size();i++){	 
    bool bError = false;

    int DepaddedStart = mOffsets[i]; // Offset relative to Synthetic, which is depadded. 
    int DepaddedEnd = mOffsets[i] + mAlignedReads[i].size() -1;
	 
    // Padded is in the absolute coordinates of Source
    int PaddedStart = mUnionSource.Depadded2Padded(DepaddedStart);// 0 based 
    int PaddedEnd = mUnionSource.Depadded2Padded(DepaddedEnd); // 0 based 

    // KJD Is this right???
    // Need to change offset to be the mUnionSource coordinate. 
    // in order for things to line up. 
    mOffsets[i] = PaddedStart+1;
		

    int SourceIdx = PaddedStart;
    int SourceIdxDepad = mUnionSource.Padded2Depadded(SourceIdx);

    Sequence SeqOut;
    SeqOut.mName = mAlignedReads[i].mName;
    SeqOut.mDescription = mAlignedReads[i].mDescription;
    for(int j = 0;j < mAlignedReads[i].size();j++){

      int BaseLoc = mBaseLocs[i][j];
      //      cout<<"BaseLoc="<<BaseLoc<<"\tSourceIdx="<<SourceIdx;
      //cout<<"\tSourceIdxDepad="<<SourceIdxDepad<<endl;

      if (BaseLoc == SourceIdxDepad){
	SeqOut.push_back(mAlignedReads[i][j]);
	SourceIdx++;
	SourceIdxDepad = mUnionSource.Padded2Depadded(SourceIdx);
      }else{
	// Scan through block of extra pads. 
	int bugCount = 0;
	bool bugSkipRead = false;
	while(BaseLoc != SourceIdxDepad){
	  SeqOut.push_back('-');
	  SourceIdx++;
	  SourceIdxDepad = mUnionSource.Padded2Depadded(SourceIdx);
	  bugCount++;
	  if (bugCount > mUnionSource.size()) break;
	  //	    cout<<"\nERROR bugCount="<<bugCount<<endl;
	  //cout<<"SeqOut.mName="<<SeqOut.mName<<" size="<<SeqOut.size()<<endl;
	  //cout<<"mUnionSource.size="<<mUnionSource.size()<<endl;
	  //cout<<"BaseLoc = "<<BaseLoc<<endl;
	  //cout<<"SourceIdxDepad = "<<SourceIdxDepad<<endl;
	  //bugSkipRead = true;
	  //break;
	  // }
	}

	if (!bugSkipRead){		  
	  // Now that they match, add the base and increment... 
	  SeqOut.push_back(mAlignedReads[i][j]);
	  SourceIdx++;
	  SourceIdxDepad = mUnionSource.Padded2Depadded(SourceIdx);
	}
      }		


    }
    if (!bError){
      mMergedReads.push_back(SeqOut);
    }
  }
}

/*

In:

ACCGTTCATCATGCAGACATACCGATTCATCCTATTTTGGGGAATTACATTTTTTTTTCATGGAC     Source
ACCGTTCATCATTGCAGACATACCGATTCATCCTATTTTTTGGGGAATTACATTTTTTTTTCATGGAC  tr1
ACCGTTCATCATGGCAGACATACCGATTCATCCTATTTTTTTGGGGAATTACATTTTTTTTTCATGGAC tr2
GTCCATGAAAAAAAAATGTAATTCCCCAAAATAGCATGAATCGGTATGTCTGCATGATGAACGGT     tr3



Out:

ACCGTTCATCAT-G-CAGACATACCGATTCATCCTATTTT---GGGGAATTACATTTTTTTTTCATGGAC USource
ACCGTTCATCATTG-CAGACATACCGATTCATCCTATTTTTT-GGGGAATTACATTTTTTTTTCATGGAC tr1
ACCGTTCATCAT-GGCAGACATACCGATTCATCCTATTTTTTTGGGGAATTACATTTTTTTTTCATGGAC tr2 
ACCGTTCATCATGCAGACATACCGATTCATGCTATTTTGGGGAATTACATTTTTTTTTCATGGAC      tr3(-)


Should be:

ACCGTTCATCAT-G-CAGACATACCGATTCATCCTATTTT---GGGGAATTACATTTTTTTTTCATGGAC USource
v
ACCGTTCATCAT-G-CAGACATACCGATTCATGCTATTTT---GGGGAATTACATTTTTTTTTCATGGAC tr3(-)

*/













/*



ACCGTTCATCATGCAGACATACCGATTCATCCTATTTTTTTTGGGGAATTACATTTTTTTTTTCATGGAC  Source


The original match to the Source will look like:

1         2         3         4         5         6         7
12345678990123456789012345678901234567890123456789012345678901234567890

ACCGTTCAT-CATGCAGACATACCGATTCATCCTATTTTTTTTGGGGAATTACATTTTTTTTTTCATGGAC Source
ACCGTTCATTCATGCAGACATACGAATTCAT-CTATTTTTTTTGGGGAATTACATTTTTTTTTTCATGGAC tr1
12345678990123456789012345678901234567890123456789012345678901234567890

And like: 


ACCGTTCATCATGCAGACATACCGATTCATCCTATTTTTTTTGGGGAATTACATTTTTTTTTTCATGGAC Source
ACCGTTCATCATGCAGACATACGAATTCATCCTATTTTTTTTGGGGAATTACATTTTTTTTT-CATGGAC tr2 
1234567890123456789012345678901234567890123456789012345678901234567890
1         2         3         4         5         6         7

Now, create a source that has the union of the pads induced by other reads:


1         2         3         4         5         6         7
12345678990123456789012345678901234567890123456789012345678901234567890
ACCGTTCAT-CATGCAGACATACCGATTCATCCTATTTTTTTTGGGGAATTACATTTTTTTTTTCATGGAC USource


So the next step is to match each read base to the consensus position 
number it had originally above:

ACCGTTCATTCATGCAGACATACGAATTCAT-CTATTTTTTTTGGGGAATTACATTTTTTTTTTCATGGAC tr1
12345678990123456789012345678901234567890123456789012345678901234567890

ACCGTTCAT CATGCAGACATACGAATTCATCCTATTTTTTTTGGGGAATTACATTTTTTTTT-CATGGAC tr2 
123456789 0123456789012345678901234567890123456789012345678901234567890



To give the merged alignment:


ACCGTTCAT-CATGCAGACATACCGATTCATCCTATTTTTTTTGGGGAATTACATTTTTTTTTTCATGGAC USource
ACCGTTCATTCATGCAGACATACGAATTCAT-CTATTTTTTTTGGGGAATTACATTTTTTTTTTCATGGAC tr1
ACCGTTCAT-CATGCAGACATACGAATTCATCCTATTTTTTTTGGGGAATTACATTTTTTTTT-CATGGAC tr2 

*/



// The original match to the SyntheticConsensus will look like:
// 
// 	   12234567890123
// Con   AC*CGTTCATCATG
// R1    ACTC*TTCATCAT
//
// Each base in R1 has a Depadded coordinate relative to Con. 
//
// Now, once we've created a Union consensus with all of the 
// read pads, we have something like:
// 
//      12233345678990123
// UCon AC*C**GTTCAT*CATG
//
// So, the next step is to match each read base to the consensus
// position number it had originally above. We match depadded coordinate 
// to depadded coordinate.  That is, the depadded coordinate for each base 
// should be the same after expanding some of the pads runs in UnionCon. 
// For example:
//
//      12233345678990123   Depadded SU Coordinates 
//      01234567890123456   Padded SU Coordinates 
// UCon AC*C**GTTCAT*CATG
// R1   ACTC  *TTCAT CAT
// 
//      1223  456789 012    BaseLocs
//
//      0123456789012345    readIdx
//
